!=========================================================================
! Copyright (C) GemTalk Systems 1986-2015.  All Rights Reserved.
!
! $Id: collisionbucket.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   CollisionBucket, AbstractCollisionBucket, Array, SequenceableCollection,
!   Collection, Object.
!
!=========================================================================

removeallmethods CollisionBucket
removeallclassmethods CollisionBucket

category: 'For Documentation Installation only'
classmethod: CollisionBucket
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A CollisionBucket is an Array that is used in a KeyValueDictionary to store
 a collection of key/value pairs for which the keys hash to the same value.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'An AbstractDictionary.  For GemStone internal use.' .
doc documentInstVar: #keyValueDictionary with: txt.

self description: doc.
%

category 'Accessing'
method: CollisionBucket
keyValueDictionary

"Returns the value of the instance variable."

^keyValueDictionary
%

category 'Updating'
method: CollisionBucket
keyValueDictionary: aDict

"Updates the value of the keyValueDictionary instance variable."

keyValueDictionary := aDict
%

category: 'Private'
method: CollisionBucket
_removeAll

"Dereferences the receiver from its parent and shrinks the receiver.
 Used while rebuilding a KeyValueDictionary."

self nilOutAll .
self size: 0
%

category: 'Private'
method: CollisionBucket
auditFor: dict with: badBucketsSet

"Audits the receiver and verifies the keyValueDictionary and
 numElements instance variables are correct.  Scans the
 receiver to ensure the number of elements it contains
 matches numElements.

 Returns the actual number of elements found in the 
 receiver.  If the audit fails, add the receiver to the
 badBucketSet argument, which is expected to be an
 IdentitySet."

|totalSize|
totalSize := 0.
1 to: self _basicSize by: 2 do:[ :idx|
  (self _at: idx) == nil
    ifFalse:[ totalSize := totalSize + 1 ].
].

keyValueDictionary == dict
  ifFalse:[badBucketsSet add: self].
totalSize = numElements
  ifFalse:[badBucketsSet add: self].

^totalSize
%
